﻿
using System.Collections.Generic;
using System.Linq;
using QuickGraph.Algorithms.RandomWalks;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterScenario;

namespace StudyCrafter.Assistant.Metrics
{
    public abstract class SimulatedPlaytraceMetric : IScenarioMetric
    {

        protected IGraphStore Store
        {
            get;
            private set;
        }

        private System.Random random = new System.Random();
        protected System.Random Rand
        {
            get
            {
                return random;
            }
            set
            {
                random = value;
            }
        }

        double eventTransitionProbability = 0;
        protected virtual double EventTransitionProbability
        {
            get
            {
                return eventTransitionProbability;
            }
            set
            {
                eventTransitionProbability = value;
            }
        }

        int maxPlaythroughLength = 1000;
        protected virtual int MaxPlaythroughLength
        {
            get
            {
                return maxPlaythroughLength;
            }
            set
            {
                maxPlaythroughLength = value;
            }
        }

        int numPlaythroughs = 500;

        public event CalculationFinishedEvent CalculationFinished;

        protected virtual int NumPlaythroughs
        {
            get
            {
                return numPlaythroughs;
            }
            set
            {
                numPlaythroughs = value;
            }
        }

        double result;
        protected virtual double Result
        {
            get
            {
                return result;
            }
            set
            {
                result = value;
            }
        }

        protected InteractionMap CurrentGraph
        {
            get; private set;
        }

        protected SimulatedPlaytraceMetric(IGraphStore store)
        {
            Store = store;
        }

        public void Calculate(Scenario scenario)
        {
            for (int i = 0; i < NumPlaythroughs; ++i)
            {
                OnStartPlaythrough(i);

                RunPlaythrough(scenario);

                OnCompletePlaythrough(i);
            }

            OnCalculationFinished();
        }

        protected virtual void OnCalculationFinished()
        {
            if (CalculationFinished != null)
            {
                CalculationFinished(Result);
            }
        }

        private void RunPlaythrough(Scenario scenario)
        {
            SceneGraph sceneGraph = scenario.GetSceneGraph(Store);

            Scene currScene = sceneGraph.Root;
            while (currScene != null)
            {
                InteractionMap interactionMap = currScene.GetInteractionMap(Store);
                Scene nextScene = DoScenePlaythrough(sceneGraph, currScene, interactionMap);
                OnSceneComplete(currScene);
                currScene = nextScene;
            }
        }

        private Scene DoScenePlaythrough(SceneGraph sceneGraph, Scene scene, InteractionMap graph)
        {
            InteractionMap preprocessedGraph = PreProcess(graph);
            CurrentGraph = preprocessedGraph;

            RandomWalkAlgorithm<InteractionNode, InteractionEdge> randomWalk = ConstructRandomWalk(preprocessedGraph);

            InteractionNode startVertex = preprocessedGraph.GetStart();
            randomWalk.StartVertex += OnStartVertex;

            Scene nextScene = null;
            randomWalk.EndVertex += ((node) =>
            {
                string nextSceneID = OnEndVertex(scene, node);

                if (!sceneGraph.TryGetVertex(new Scene(nextSceneID), out nextScene))
                {
                    nextScene = null;
                }
            });

            randomWalk.TreeEdge += OnTreeEdge;

            randomWalk.Generate(startVertex, MaxPlaythroughLength);

            return nextScene;
        }

        private InteractionMap PreProcess(InteractionMap graph)
        {
            InteractionMap clonedGraph = graph.Clone();
            return AddRandomEventNodeEdges(clonedGraph);
        }

        private InteractionMap AddRandomEventNodeEdges(InteractionMap graph)
        {
			int maxEdgeID = graph.Graph.Edges.Select(edge => edge.ID).Max();
			IEdgeIDGenerator<InteractionNode, InteractionEdge> edgeIDGen = new SequentialEdgeIDGenerator<InteractionNode, InteractionEdge>(maxEdgeID + 1);

            List<InteractionNode> events = graph.FindEventBasedInteractionNodes().ToList();
            if (events.Count == 0)
            {
                // There are no events, so we don't need to add any transitions to them.
                return graph;
            }

            foreach (InteractionNode node in graph.Graph.Vertices)
            {
                if (node.NodeType == InteractionNode.Type.End)
                {
                    // No edges out of end nodes.
                    continue;
                }

                if (Rand.NextDouble() < EventTransitionProbability)
                {
                    InteractionNode randomEvent = events[Rand.Next(events.Count - 1)];
					InteractionEdge newEdge = new InteractionEdge(InteractionEdge.Type.Indirect, node, randomEvent);
					newEdge.SetID(edgeIDGen.GenerateID(newEdge));
                    graph.Graph.AddEdge(newEdge);
                }
            }

            return graph;
        }

        protected virtual RandomWalkAlgorithm<InteractionNode, InteractionEdge> ConstructRandomWalk(InteractionMap graph)
        {
            return new RandomWalkAlgorithm<InteractionNode, InteractionEdge>(graph.Graph);
        }

        protected string DetermineNextScene(ScriptGraph scriptGraph, InteractionNode node)
        {
            ScriptNode scriptNode;
            if (!scriptGraph.TryGetVertex(new ScriptNode(node.ID), out scriptNode))
            {
                return null;
            }

            StringProperty nextSceneProperty;
            if (scriptNode.Type != ScriptNodeTypes.End
                || !scriptNode.TryGetProperty("nextScript", out nextSceneProperty))
            {
                return null;
            }

            return nextSceneProperty.Val;
        }

        protected virtual void OnStartVertex(InteractionNode node)
        {

        }

        protected virtual string OnEndVertex(Scene scene, InteractionNode node)
        {
            return DetermineNextScene(scene.GetScriptGraph(Store), node);
        }

        protected virtual void OnTreeEdge(InteractionEdge edge)
        {

        }

        protected virtual void OnStartPlaythrough(int playthroughNum)
        {
        }

        protected virtual void OnCompletePlaythrough(int playthroughNum)
        {
        }

        protected virtual void OnSceneComplete(Scene scene)
        {
        }
    }
}
