﻿
using System;
using System.Linq;
using NUnit.Framework;
using Sprache;

namespace StudyCrafter.Assistant.Representation
{
    public class ExpressionParsingTest
    {
        [Test]
        public void TestIdentifier()
        {
            Assert.AreEqual(
                "{Test}", (ExpressionParser.Identifier).Parse("{Test}").Val);

            Assert.Throws(typeof(Sprache.ParseException), () =>
                 (ExpressionParser.Identifier).Parse("Test"));

            Assert.Throws(typeof(Sprache.ParseException), () =>
                          (ExpressionParser.Identifier).Parse("{Test"));

            Assert.Throws(typeof(Sprache.ParseException), () =>
                          (ExpressionParser.Identifier).Parse("Test}"));
        }

        [Test]
        public void TestDouble()
        {
            Assert.AreEqual(
                 0.01, (ExpressionParser.DoubleValue).Parse("0.01").DoubleValue);

            Assert.Throws(typeof(Sprache.ParseException), () =>
                 (ExpressionParser.DoubleValue).Parse("Test"));
        }

        [Test]
        public void TestInt()
        {
            Assert.AreEqual(
                1, (ExpressionParser.IntValue).Parse("1").IntValue);

            Assert.Throws(typeof(Sprache.ParseException), () =>
                 (ExpressionParser.IntValue).Parse("0.01"));

            Assert.Throws(typeof(Sprache.ParseException), () =>
                          (ExpressionParser.IntValue).Parse("Test"));
        }


        [Test]
        public void TestBool()
        {
            Assert.AreEqual(
                true, (ExpressionParser.BoolValue).Parse("True").BoolValue);

            Assert.AreEqual(
                false, (ExpressionParser.BoolValue).Parse("False").BoolValue);

            Assert.Throws(typeof(Sprache.ParseException), () =>
                 (ExpressionParser.DoubleValue).Parse("Test"));
        }

        [Test]
        public void TestTerminal()
        {
            Assert.DoesNotThrow(() => ExpressionParser.LowPriorityTerminal.Parse("0.01"));
            Assert.DoesNotThrow(() => ExpressionParser.LowPriorityTerminal.Parse("1"));
            Assert.DoesNotThrow(() => ExpressionParser.LowPriorityTerminal.Parse("False"));
            Assert.DoesNotThrow(() => ExpressionParser.LowPriorityTerminal.Parse("True"));
            Assert.DoesNotThrow(() => ExpressionParser.LowPriorityTerminal.Parse("{Test}"));

            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.LowPriorityTerminal.Parse("(5)"));
        }

        [Test]
        public void TestTerm()
        {
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("0.01"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("(0.01)"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("1"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("(1)"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("False"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("(False)"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("True"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("(True)"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("{Test}"));
            Assert.DoesNotThrow(() => ExpressionParser.Term.Parse("({Test})"));

            Assert.IsTrue(ExpressionParser.Term.Parse("0.01").IsTerminal);
            Assert.AreEqual("0.01", ExpressionParser.Term.Parse("0.01").Terminal.Val);
            Assert.AreEqual("double", ExpressionParser.Term.Parse("0.01").Terminal.ID);
            Assert.AreEqual(0.01, (ExpressionParser.Term.Parse("0.01").Terminal as DoubleProperty).DoubleValue);

            Assert.IsTrue(ExpressionParser.Term.Parse("1").IsTerminal);
            Assert.AreEqual("1", ExpressionParser.Term.Parse("1").Terminal.Val);
            Assert.AreEqual("int", ExpressionParser.Term.Parse("1").Terminal.ID);
            Assert.AreEqual(1, (ExpressionParser.Term.Parse("1").Terminal as IntProperty).IntValue);

            Assert.IsTrue(ExpressionParser.Term.Parse("True").IsTerminal);
            Assert.AreEqual("True", ExpressionParser.Term.Parse("True").Terminal.Val);
            Assert.AreEqual("bool", ExpressionParser.Term.Parse("True").Terminal.ID);
            Assert.AreEqual(true, (ExpressionParser.Term.Parse("True").Terminal as BoolProperty).BoolValue);

            Assert.IsTrue(ExpressionParser.Term.Parse("False").IsTerminal);
            Assert.AreEqual("False", ExpressionParser.Term.Parse("False").Terminal.Val);
            Assert.AreEqual("bool", ExpressionParser.Term.Parse("False").Terminal.ID);
            Assert.AreEqual(false, (ExpressionParser.Term.Parse("False").Terminal as BoolProperty).BoolValue);
        }

        [Test]
        public void TestOperand()
        {
            Assert.DoesNotThrow(() => ExpressionParser.Operand.Parse("0.01"));
            Assert.DoesNotThrow(() => ExpressionParser.Operand.Parse("-0.01"));
            Assert.DoesNotThrow(() => ExpressionParser.Operand.Parse("-(0.01)"));
        }

        [Test]
        public void TestSubExprs()
        {
            Assert.AreEqual(OpType.Power, ExpressionParser.MidPrioritySubExpr.End().Parse("1^3").Operator.Type);
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.HighPrioritySubExpr.End<Operand>().Parse("1*3"));

            Assert.AreEqual(OpType.Multiply, ExpressionParser.MidPrioritySubExpr.End().Parse("1*3").Operator.Type);
            Assert.AreEqual(OpType.Divide, ExpressionParser.MidPrioritySubExpr.End().Parse("1/3").Operator.Type);
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.MidPrioritySubExpr.End<Operand>().Parse("1+3"));

            Assert.AreEqual(OpType.Subtract, ExpressionParser.LowPrioritySubExpr.End().Parse("1-3").Operator.Type);
            Assert.AreEqual(OpType.Add, ExpressionParser.LowPrioritySubExpr.End().Parse("1+3").Operator.Type);
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.LowPrioritySubExpr.End<Operand>().Parse("1>3"));

            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.LowPrioritySubExpr.End<Operand>().Parse("{Test}?"));
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.LowPrioritySubExpr.End<Operand>().Parse("T"));
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.LowPrioritySubExpr.End<Operand>().Parse("-"));

            // Spaces
            Assert.AreEqual(OpType.Power, ExpressionParser.MidPrioritySubExpr.End().Parse("1^ 3").Operator.Type);
            Assert.AreEqual(OpType.Power, ExpressionParser.MidPrioritySubExpr.End().Parse("1 ^3").Operator.Type);
            Assert.AreEqual(OpType.Power, ExpressionParser.MidPrioritySubExpr.End().Parse("1 ^ 3").Operator.Type);
        }

        [Test]
        public void TestComparisons()
        {
            Assert.AreEqual(OpType.Equals, ExpressionParser.Expression.End().Parse("1==1").Operator.Type);
            Assert.AreEqual(OpType.Less, ExpressionParser.Expression.End().Parse("1<1").Operator.Type);
            Assert.AreEqual(OpType.Greater, ExpressionParser.Expression.End().Parse("1>1").Operator.Type);
            Assert.AreEqual(OpType.Greater_OR, ExpressionParser.Expression.End().Parse("1>=1").Operator.Type);
            Assert.AreEqual(OpType.Less_OR, ExpressionParser.Expression.End().Parse("1<=1").Operator.Type);

            // Spaces
            Assert.AreEqual(OpType.Greater, ExpressionParser.Expression.End().Parse("1> 1").Operator.Type);
            Assert.AreEqual(OpType.Greater, ExpressionParser.Expression.End().Parse("1 >1").Operator.Type);
            Assert.AreEqual(OpType.Greater, ExpressionParser.Expression.End().Parse("1 > 1").Operator.Type);

            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.Expression.End<Operand>().Parse("{Test}?"));
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.Expression.End<Operand>().Parse("T"));
            Assert.Throws(typeof(Sprache.ParseException), () => ExpressionParser.Expression.End<Operand>().Parse("-"));
        }

        [Test]
        public void TestComplexNumericExpressions()
        {
            Operand result = ExpressionParser.Expression.End().Parse("(1+ 2.0) ==3");
            Assert.AreEqual(OpType.Equals, result.Operator.Type);
            Assert.AreEqual(OpType.Add, result.Operator.Operands.First().Operator.Type);
            Assert.AreEqual(3, (result.Operator.Operands.ElementAt(1).Terminal as IntProperty).IntValue);
            Assert.AreEqual(1, (result.Operator.Operands.First().Operator.Operands.First().Terminal as IntProperty).IntValue);
            Assert.AreEqual(2.0, (result.Operator.Operands.First().Operator.Operands.ElementAt(1).Terminal as DoubleProperty).DoubleValue);
        }

        [Test]
        public void TestComplexBoolExpressions()
        {
            Operand result = ExpressionParser.Expression.End().Parse("(True ) !=false");
            Assert.AreEqual(OpType.NotEqual, result.Operator.Type);
            Assert.AreEqual(true, (result.Operator.Operands.First().Terminal as BoolProperty).BoolValue);
            Assert.AreEqual(false, (result.Operator.Operands.ElementAt(1).Terminal as BoolProperty).BoolValue);
        }

        [Test]
        public void TestStringExpression()
        {
            Operand result = ExpressionParser.ExpressionOrStringExpression.End().Parse("This is a string using a {variable} or {two}.");
            Assert.IsNotNull(result.Terminal as StringExpressionProperty);
            Assert.AreEqual("variable", (result.Terminal as StringExpressionProperty).Variables.Get<VariableProperty>("0").Variable);
            Assert.AreEqual("two", (result.Terminal as StringExpressionProperty).Variables.Get<VariableProperty>("1").Variable);
        }
    }
}