﻿using System;
using System.Linq;
using Sprache;

namespace StudyCrafter.Assistant.Representation
{
    // Expression Grammar:
    //
    // ExpressionOrStringExpression := Expression | StringExpression
    // Expression := Comparison | LowPrioritySubExpr
    // Comparison := LowPrioritySubExpr ">"|"<"|"=="|">="|"<="|"!=" LowPrioritySubExpr
    // LowPrioritySubExpr := MidPrioritySubExpr ["+"|"-" MidPrioritySubExpr]*
    // MidPrioritySubExpr := HighPrioritySubExpr ["*"|"/"|"%" HighPrioritySubExpr]*
    // HighPrioritySubExpr := Operand ["^" Operand]*
    // Operand := "-"Term | Term
    // Term := Terminal | "("Expression")"
    // Terminal := Identifier | <a number> | <a boolean constant>
    // Identifier := "{StringLiteral}"
    // StringLiteral := <any string>
    //
    // StringExpressionComponent := StringLiteral|Identifier
    // StringExpression := StringExpressionComponent*

    public static class ExpressionParser
    {
        public static ExpressionProperty FromString(this ExpressionProperty property, string text)
        {
            IOperand fullExpression = ExpressionOrStringExpression.Parse(text);
            property.Root = fullExpression;
            return property;
        }

        public static Parser<VariableProperty> Identifier
        {
            get
            {
                return
                    from name in Variable
                    select new VariableProperty("variable", name);
            }
        }

        public static Parser<DoubleProperty> DoubleValue
        {
            get
            {
                return Parse.Decimal.Select(x => new DoubleProperty("double", x));
            }
        }

        public static Parser<IntProperty> IntValue
        {
            get
            {
                return
                    from number in Parse.Number
                    from notPeriod in Parse.Not(Parse.Char('.'))
                    select new IntProperty("int", number);
            }
        }

        public static Parser<BoolProperty> BoolValue
        {
            get
            {
                return (Parse.IgnoreCase("false").Token().Or(Parse.IgnoreCase("true").Token())).Text().Select(x => new BoolProperty("bool", x));
            }
        }

        public static Parser<Operand> Terminal
        {
            get
            {
                return Identifier.Select(property => (new TerminalOperand(property) as Operand))
                                 .Or<Operand>(IntValue.Select(property => new TerminalOperand(property) as Operand))
                                 .Or<Operand>(BoolValue.Select(property => new TerminalOperand(property) as Operand));
            }
        }

        public static Parser<Operand> LowPriorityTerminal
        {
            get
            {
                return Terminal.Or<Operand>(DoubleValue.Select(property => new TerminalOperand(property) as Operand));
            }
        }

        public static Parser<Operand> Term
        {
            get
            {
                return
                    (from parentheses in Parse.Char('(')
                     from expr in Expression
                     from closeParentheses in Parse.Char(')')
                     select expr)
                    .Or<Operand>(
                            LowPriorityTerminal
                    );
            }
        }

        public static Parser<Operand> Operand
        {
            get
            {
                return
                    (from negateSign in Parse.Char('-')
                     from term in Term
                     select new OperatorOperand(new NegateOperator(term)) as Operand)
                        .Or<Operand>(Term);
            }
        }

        public static Parser<OpType> Operator(string text, OpType type)
        {
            return Parse.String(text).Token().Return(type);
        }

        // Binary operators
        public static Parser<OpType> Add = Operator("+", OpType.Add);
        public static Parser<OpType> Subtract = Operator("-", OpType.Subtract);
        public static Parser<OpType> Multiply = Operator("*", OpType.Multiply);
        public static Parser<OpType> Divide = Operator("/", OpType.Divide);
        public static Parser<OpType> Power = Operator("^", OpType.Power);

        // Comparisons
        public static Parser<OpType> Less = Operator("<", OpType.Less);
        public static Parser<OpType> Greater = Operator(">", OpType.Greater);
        public static Parser<OpType> Equal = Operator("=", OpType.Equals);
        public static Parser<OpType> DoubleEqual = Operator("==", OpType.Equals);
        public static Parser<OpType> Greater_OR = Operator(">=", OpType.Greater_OR);
        public static Parser<OpType> Less_OR = Operator("<=", OpType.Less_OR);
        public static Parser<OpType> NotEqual = Operator("!=", OpType.NotEqual);

        public static Parser<Operand> HighPrioritySubExpr
        {
            get
            {
                return Parse.ChainOperator(Power, Operand,
                                           (opType, left, right) => new OperatorOperand(new BinaryOperator(left, opType, right)) as Operand);
            }
        }

        public static Parser<Operand> MidPrioritySubExpr
        {
            get
            {
                return Parse.ChainOperator(Multiply.XOr(Divide), HighPrioritySubExpr,
                                                             (opType, left, right) => new OperatorOperand(new BinaryOperator(left, opType, right)) as Operand);
            }
        }

        public static Parser<Operand> LowPrioritySubExpr
        {
            get
            {
                return Parse.ChainOperator(Add.XOr(Subtract), MidPrioritySubExpr,
                                                             (opType, left, right) => new OperatorOperand(new BinaryOperator(left, opType, right)) as Operand);
            }
        }

        public static Parser<OpType> Comparison
        {
            get
            {
                return Less_OR.Or(Less).Or(Greater_OR).Or(Greater).Or(DoubleEqual).Or(Equal).Or(NotEqual);
            }
        }

        public static Parser<Operand> Expression
        {
            get
            {
                return Parse.ChainOperator(Comparison, LowPrioritySubExpr,
                                           (comparisonType, left, right) => new OperatorOperand(new BinaryOperator(left, comparisonType, right)) as Operand);
            }
        }

        public static Parser<string> StringLiteral
        {
            get
            {
                return from text in Parse.CharExcept('{').AtLeastOnce().Text()
                       select text;
            }
        }

        public static Parser<string> Variable
        {
            get
            {
                return
                    from beginChar in Parse.Char('{').Once().Text()
                    from name in Parse.CharExcept('}').AtLeastOnce().Token().Text()
                    from endChar in Parse.Char('}').Once().Text()
                    select beginChar + name + endChar;
            }
        }

        public static Parser<string> StringExpressionComponent
        {
            get
            {
                return StringLiteral.Or(Variable);
            }
        }

        public static Parser<StringExpressionProperty> StringExpression
        {
            get
            {
                return StringExpressionComponent.Many().Select((strings) => new StringExpressionProperty("string", String.Concat(strings.ToArray())));
            }
        }

        public static Parser<Operand> ExpressionOrStringExpression
        {
            get
            {
                return Expression.End().Or(StringExpression.End().Select((expression) => new TerminalOperand(expression) as Operand));
            }
        }

    }
}