﻿using System;
using System.Collections.Generic;
using System.Linq;
using QuickGraph;
using QuickGraph.Algorithms;
using QuickGraph.Algorithms.Search;
using QuickGraph.Algorithms.Services;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterScenario;

namespace StudyCrafter.Assistant.StudyCrafterLoading
{
	// TODO:
	//
	// - Figure out whether to coalesce different types of feedback further?
	// - Determine what to do about branches that are not clearly random.
	// - Determine whether/how to do indirect edges.
	// - Determine whether we are missing cases or nuances for event types.
	// - What do we do about "open response" nodes?
	// - Determine whether there are more special nodes types to handle.
	// - Add additional metadata on the nodes depending on the script? e.g. what type of event? is it a button press?

	public class ScriptToInteractionMapAlgorithm : AlgorithmBase<IBidirectionalGraph<ScriptNode, IEdge<ScriptNode>>>
	{
		private ScriptToInteractionMapping scriptToInteractionMap = new ScriptToInteractionMapping();
		private IEdgeIDGenerator<InteractionNode, InteractionEdge> edgeIDGen;
		private IObjectIDGenerator<InteractionNode> nodeIDGen;

		public InteractionMap ResultMap
		{
			get; private set;
		}

		public ScriptToInteractionMapAlgorithm(InteractionMap interactionMapToFill, ScriptGraph visitedGraph,
											   IEdgeIDGenerator<InteractionNode, InteractionEdge> edgeIDGen = null,
											   IObjectIDGenerator<InteractionNode> nodeIDGen = null) : base(visitedGraph.Graph)
		{
			this.edgeIDGen = edgeIDGen;
			if (null == this.edgeIDGen)
			{
				this.edgeIDGen = new SequentialEdgeIDGenerator<InteractionNode, InteractionEdge>();
			}

			this.nodeIDGen = nodeIDGen;
			if (null == this.nodeIDGen)
			{
				this.nodeIDGen = new SequentialObjectIDGenerator<InteractionNode>();
			}

			ResultMap = interactionMapToFill;
		}

		public ScriptToInteractionMapAlgorithm(InteractionMap interactionMapToFill, IAlgorithmComponent host, ScriptGraph visitedGraph,
											   IEdgeIDGenerator<InteractionNode, InteractionEdge> edgeIDGen = null,
											   IObjectIDGenerator<InteractionNode> nodeIDGen = null) : base(host, visitedGraph.Graph)
		{
			this.edgeIDGen = edgeIDGen;
			if (null == this.edgeIDGen)
			{
				this.edgeIDGen = new SequentialEdgeIDGenerator<InteractionNode, InteractionEdge>();
			}

			this.nodeIDGen = nodeIDGen;
			if (null == this.nodeIDGen)
			{
				this.nodeIDGen = new SequentialObjectIDGenerator<InteractionNode>();
			}

			ResultMap = interactionMapToFill;
		}

		protected override void InternalCompute()
		{
			DepthFirstSearchAlgorithm<ScriptNode, IEdge<ScriptNode>> dfs =
				new DepthFirstSearchAlgorithm<ScriptNode, IEdge<ScriptNode>>(this, VisitedGraph, new Dictionary<ScriptNode, GraphColor>(VisitedGraph.VertexCount));
			dfs.StartVertex += this.OnStartVertex;
			dfs.TreeEdge += this.OnTreeEdge;
			dfs.BackEdge += this.OnBackEdge;
			dfs.ForwardOrCrossEdge += this.OnForwardOrCrossEdge;
			dfs.Finished += this.PostProc;
			dfs.Compute();

			// FIXME: Does this work with async operations? Or will we return too early?
		}

		private void PostProc(object sender, EventArgs args)
		{
			// Clean up the interaction graph and remove any non-interaction nodes.
			List<InteractionNode> allNodes = new List<InteractionNode>(ResultMap.Graph.Vertices);
			foreach (InteractionNode node in allNodes)
			{
				if (node.NodeType == InteractionNode.Type.NonInteraction)
				{
					RemoveNode(node);
				}
			}

			InteractionMapPostProcAlgorithm postProc =
				new InteractionMapPostProcAlgorithm(ResultMap);

			postProc.Compute();
		}

		private void RemoveNode(InteractionNode node)
		{
			EdgeFactory<InteractionNode, InteractionEdge> edgeFactory = new EdgeFactory<InteractionNode, InteractionEdge>(CreateEdge);
			((BidirectionalGraph<InteractionNode, InteractionEdge>)ResultMap.Graph).MergeVertex(node, edgeFactory);
		}

		private InteractionEdge CreateEdge(InteractionNode first, InteractionNode second)
		{
			InteractionEdge edge;
			if (first.NodeType == InteractionNode.Type.InteractionPoint)
			{
				edge = new InteractionEdge(InteractionEdge.Type.Choice, first, second);
			}
			else
			{
				edge = new InteractionEdge(InteractionEdge.Type.Direct, first, second);
			}

			edge.SetID(GenerateEdgeID(edge));
			return edge;
		}

		private void OnStartVertex(ScriptNode scriptNode)
		{
			InteractionNode newNode;

			if (!scriptToInteractionMap.TryGetEntryPoint(scriptNode, out newNode))
			{
				newNode = BuildInteractionSubgraph(scriptNode);
			}

			if (scriptNode.Type.Equals(ScriptNodeTypes.CustomFunction))
			{
				BuildCustomFunctionEdges(scriptNode, newNode);
			}
		}

		private void BuildCustomFunctionEdges(ScriptNode scriptNode, InteractionNode customFunctionEntry)
		{
			// Find call points for custom function.
			IEnumerable<ScriptNode> callPoints = VisitedGraph.Vertices.Where(
				(vertex) => (vertex.Type.Equals(ScriptNodeTypes.CallFunction)
							 && vertex.GetProperty<StringProperty>("functionType").Val.Equals("CustomFunction"))
							 && vertex.GetProperty<StringProperty>("functionDetail").Val.Equals(scriptNode.GetProperty<StringProperty>("functionName").Val)
			);

			// For any nodes in the set of call points that have already been created, build new edges.
			foreach (ScriptNode callPoint in callPoints)
			{
				InteractionNode callPointExitNode;
				if (scriptToInteractionMap.TryGetExitPoint(callPoint, out callPointExitNode))
				{
					InteractionEdge newEdge = CreateDirectEdge(callPointExitNode, customFunctionEntry);
					ResultMap.Graph.AddEdge(newEdge);
				}
			}
		}

		private void BuildCallFunctionEdges(ScriptNode scriptNode, InteractionNode interactionNode)
		{
			if (!scriptNode.GetProperty<StringProperty>("functionType").Val.Equals("CustomFunction"))
			{
				// If this is not a custom function, we don't know how to connect it.
				// TODO: Should we do something about some of these?
				return;
			}

			// Find called custom function.
			IEnumerable<ScriptNode> customFunctionNodes = VisitedGraph.Vertices.Where(
				(vertex) => (vertex.Type.Equals(ScriptNodeTypes.CustomFunction)
							 && vertex.GetProperty<StringProperty>("functionName").Val.Equals(scriptNode.GetProperty<StringProperty>("functionDetail").Val))
			);

			// If the custom function has already been created, associate it with an edge.
			// (If it hasn't, we'll associate them later, when the custom function node is built.)
			foreach (ScriptNode customFunctionNode in customFunctionNodes)
			{
				InteractionNode functionEntryPoint;
				if (scriptToInteractionMap.TryGetEntryPoint(customFunctionNode, out functionEntryPoint))
				{
					InteractionEdge newEdge = CreateDirectEdge(interactionNode, functionEntryPoint);
					ResultMap.Graph.AddEdge(newEdge);
				}
			}
		}

		private void OnTreeEdge(IEdge<ScriptNode> edge)
		{
			BuildNodesAndEdge(edge);
		}

		private void BuildNodesAndEdge(IEdge<ScriptNode> edge)
		{
			InteractionNode sourceInteractionNode;
			InteractionNode targetInteractionNode;

			if (!scriptToInteractionMap.TryGetExitPoint(edge.Source, out sourceInteractionNode))
			{
				sourceInteractionNode = BuildInteractionSubgraph(edge.Source);
			}

			if (!scriptToInteractionMap.TryGetEntryPoint(edge.Target, out targetInteractionNode))
			{
				targetInteractionNode = BuildInteractionSubgraph(edge.Target);
				if (null == targetInteractionNode)
				{
					// If this is not a type of interaction node, just associate the script node with the current interaction node.
					scriptToInteractionMap.RegisterSingleNodeSubgraph(edge.Target, sourceInteractionNode);

					// If the end of this edge is a "call function" node, associate it with the target of its call, 
					// if an interaction node has already been made for it (otherwise, we'll associate when it gets there).
					if (edge.Target.Type.Equals(ScriptNodeTypes.CallFunction))
					{
						BuildCallFunctionEdges(edge.Target, sourceInteractionNode);
					}
					return;
				}
			}

			InteractionEdge newEdge = CreateDirectEdge(sourceInteractionNode, targetInteractionNode);

			// If the end of this edge is a "call function" node, associate it with the target of its call, 
			// if an interaction node has already been made for it (otherwise, we'll associate when it gets there).
			if (edge.Target.Type.Equals(ScriptNodeTypes.CallFunction))
			{
				BuildCallFunctionEdges(edge.Target, targetInteractionNode);
			}

			ResultMap.Graph.AddEdge(newEdge);
		}

		private InteractionEdge CreateDirectEdge(InteractionNode source, InteractionNode target)
		{
			InteractionEdge edge;
			if (source.NodeType.Equals(InteractionNode.Type.InteractionPoint))
			{
				edge = new InteractionEdge(InteractionEdge.Type.Choice, source, target);
			}
			else
			{
				edge = new InteractionEdge(InteractionEdge.Type.Direct, source, target);
			}
			edge.SetID(GenerateEdgeID(edge));
			return edge;
		}

		private void OnBackEdge(IEdge<ScriptNode> edge)
		{
			// TODO: Do we need to treat this differently?
			BuildNodesAndEdge(edge);
		}

		private void OnForwardOrCrossEdge(IEdge<ScriptNode> edge)
		{
			// TODO: Do we need to treat this differently?
			BuildNodesAndEdge(edge);
		}

		// Builds a node or subgraph for the given scriptNode, returning the node (or entry point of the subgraph)
		private InteractionNode BuildInteractionSubgraph(ScriptNode scriptNode)
		{
			InteractionNode newNode = null;
			switch (scriptNode.Type)
			{
				case ScriptNodeTypes.Event:
					{
						StringProperty eventType;
						if (scriptNode.TryGetProperty<StringProperty>("eventType", out eventType) &&
							eventType.Val.Equals("TimePassed"))
						{
							// A time passed event is independent of user input, so we treat it as an event.
							newNode = InitAndAdd(InteractionNode.Type.Event, scriptNode);
							RegisterSingleNode(scriptNode, newNode);
						}
						else
						{
							// Other types of event happen in response to user input, so they are interaction points.
							newNode = InitAndAdd(InteractionNode.Type.InteractionPoint, scriptNode);
							RegisterSingleNode(scriptNode, newNode);
						}
						break;
					}
				case ScriptNodeTypes.End:
					{
						newNode = InitAndAdd(InteractionNode.Type.End, scriptNode);
						break;
					}
				case ScriptNodeTypes.Action:
					{
						if (scriptNode.ID.Equals("0"))
						{
							// The node with ID 0 is the start node, and it is represented as an action.
							newNode = InitAndAdd(InteractionNode.Type.Start, scriptNode);
							RegisterSingleNode(scriptNode, newNode);
						}
						else
						{
							newNode = InitAndAdd(InteractionNode.Type.Unknown, scriptNode);
							RegisterSingleNode(scriptNode, newNode);
						}
						break;
					}
				case ScriptNodeTypes.Dialogue:
				case ScriptNodeTypes.Sound:
					{
						newNode = InitAndAdd(InteractionNode.Type.Unknown, scriptNode);
						RegisterSingleNode(scriptNode, newNode);
						break;
					}
				case ScriptNodeTypes.Branch:
					{
						// TODO: is there another case where we want to treat these as events?
						if (scriptNode.GetProperty<BoolProperty>("isRandom").BoolValue)
						{
							newNode = InitAndAdd(InteractionNode.Type.Event, scriptNode);
							RegisterSingleNode(scriptNode, newNode);
						}
						else
						{
							newNode = InitAndAdd(InteractionNode.Type.NonInteraction, scriptNode);
							RegisterSingleNode(scriptNode, newNode);
						}
						break;
					}
				case ScriptNodeTypes.Choice:
					{
						// Create an interaction point.
						newNode = BuildInteractionPointFromChoice(scriptNode);
						break;
					}
				default:
					{
						newNode = InitAndAdd(InteractionNode.Type.NonInteraction, scriptNode);
						RegisterSingleNode(scriptNode, newNode);
						break;
					}
			}

			return newNode;
		}

		private InteractionNode InitAndAdd(InteractionNode.Type newNodeType, ScriptNode scriptNode)
		{
			InteractionNode newNode = new InteractionNode(newNodeType);
			GenerateVertexID(newNode);
			FillPropertiesFromScript(newNode, scriptNode);
			ResultMap.Graph.AddVertex(newNode);

			return newNode;
		}

		private void RegisterSingleNode(ScriptNode scriptNode, InteractionNode interactionNode)
		{
			scriptToInteractionMap.RegisterSingleNodeSubgraph(scriptNode, interactionNode);
		}

		private InteractionNode BuildInteractionPointFromChoice(ScriptNode scriptNode)
		{
			InteractionNode interactionPoint = InitAndAdd(InteractionNode.Type.InteractionPoint, scriptNode);
			InteractionNode resultNode;

			StringExpressionProperty dialogueProp;
			if (scriptNode.TryGetProperty<StringExpressionProperty>("content", out dialogueProp))
			{
				// This node has dialogue at the start, which should be represented as an event or feedback prior to the player's choice.

				// Create a new result node to precede it.
				resultNode = InitAndAdd(InteractionNode.Type.Unknown, scriptNode);

				InteractionEdge newEdge = CreateDirectEdge(resultNode, interactionPoint);
				ResultMap.Graph.AddEdge(newEdge);

				scriptToInteractionMap.RegisterEntryPoint(scriptNode, resultNode);
				scriptToInteractionMap.RegisterExitPoint(scriptNode, interactionPoint);
			}
			else
			{
				resultNode = interactionPoint;
				RegisterSingleNode(scriptNode, interactionPoint);
			}

			return resultNode;
		}

		private void GenerateVertexID(InteractionNode vertex)
		{
			string id = nodeIDGen.GenerateID(vertex);
			vertex.ID = id;
		}

		private void FillPropertiesFromScript(InteractionNode newNode, ScriptNode scriptNode)
		{
			newNode.SetProperty(new StringProperty("script_type", scriptNode.Type));
			newNode.SetProperty(new StringProperty("start_script_id", scriptNode.ID));
		}

		private int GenerateEdgeID(InteractionEdge edge)
		{
			return edgeIDGen.GenerateID(edge);
		}
	}
}
