﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using StudyCrafter.Assistant.Loading;
using UnityEngine;

namespace StudyCrafter.Assistant.StudyCrafterLoading
{
	[RequireComponent(typeof(StreamingAssetsFileLoader))]
	public class StreamingAssetsProjectLoader : MonoBehaviour, IProjectLoader
	{
		private class ProjectLoadData
		{
			public ProjectLoadStatus Status
			{
				get;
				private set;
			}

			public JsonScenarioStrings Strings
			{
				get;
				set;
			}

			public ProjectLoadData()
			{
				Status = new ProjectLoadStatus();
				Strings = new JsonScenarioStrings();
			}

			public void UpdateStatus(LoadStatus status)
			{
				Status.UpdateStatus(status);
			}

			public void UpdateRootJson(string rootJson)
			{
				Strings.RootJson = rootJson;
			}

			public void AddSceneJson(string sceneID, string sceneJson)
			{
				Strings.AddScene(sceneID, sceneJson);
			}

			public void NotifyError(string error)
			{
				Status.NotifyError(error);
			}
		}

		private IDictionary<string, ProjectLoadData> projectData = new Dictionary<string, ProjectLoadData>();

		private StreamingAssetsFileLoader fileLoader;

		[SerializeField]
		private string projectsFolder = "NewProjectSystem";

		public string ProjectsFolder
		{
			get
			{
				return projectsFolder;
			}
			set
			{
				projectsFolder = value;
			}
		}

		public StreamingAssetsProjectLoader()
		{
		}

		public void Start()
		{
			fileLoader = GetComponent<StreamingAssetsFileLoader>();
		}

		public JsonScenarioStrings GetLoadedProject(string projectName)
		{
			ProjectLoadData data;
			if (!projectData.TryGetValue(projectName, out data))
			{
				return null;
			}

			if (data.Status.Status != LoadStatus.Loaded)
			{
				return null;
			}

			return data.Strings;
		}

		public ProjectLoadStatus GetStatus(string projectName)
		{
			ProjectLoadData data;
			if (!projectData.TryGetValue(projectName, out data))
			{
				data = new ProjectLoadData();
				projectData.Add(projectName, data);
			}

			return data.Status;
		}

		public ProjectLoadStatus LoadProject(string projectName)
		{
			ProjectLoadData data;
			if (!projectData.TryGetValue(projectName, out data))
			{
				data = new ProjectLoadData();
				projectData.Add(projectName, data);
			}

			if (data.Status.Status == LoadStatus.InProgress)
			{
				// Already loading, don't start again.
				return data.Status;
			}

			StartCoroutine(Load(projectName));

			return data.Status;
		}

		private IEnumerator Load(string projectName)
		{
			ProjectLoadData loadData = projectData[projectName];
			loadData.UpdateStatus(LoadStatus.InProgress);

			// Load the directory index, so we know what scene files exist:
			List<string> sceneFilesToLoad = new List<string>();

			yield return DetermineSceneFiles(projectName, sceneFilesToLoad);
			if (sceneFilesToLoad.Count == 0)
			{
				loadData.NotifyError("Project not found.");
				yield break;
			}

			foreach (string sceneFile in sceneFilesToLoad)
			{
				Debug.Log("Scene file: " + sceneFile);

				string scenePath = GetScenePath(projectName, sceneFile);
				yield return LoadSceneFile(scenePath, loadData);

				string sceneString = fileLoader.RetrieveResult(scenePath);
				if (IsRoot(sceneFile))
				{
					loadData.UpdateRootJson(sceneString);
				}
				else
				{
					loadData.AddSceneJson(sceneFile, sceneString);
				}
			}

			loadData.UpdateStatus(LoadStatus.Loaded);
		}

		private bool IsRoot(string sceneName)
		{
			return sceneName.StartsWith("00_", comparisonType: System.StringComparison.InvariantCultureIgnoreCase);
		}

		private IEnumerator LoadSceneFile(string scenePath, ProjectLoadData loadData)
		{
			LoadStatus sceneFileLoadStatus = fileLoader.Load(scenePath);
			while (sceneFileLoadStatus != LoadStatus.Loaded)
			{
				yield return null;
				sceneFileLoadStatus = fileLoader.GetStatus(scenePath);
			}
		}

		private string GetScenePath(string projectName, string sceneName)
		{
			string basePath = GetProjectBasePath(projectName);
			return System.IO.Path.Combine(basePath, sceneName);
		}

		private string GetProjectBasePath(string projectName)
		{
			return System.IO.Path.Combine(projectsFolder, projectName);
		}

		private IEnumerator DetermineSceneFiles(string projectName, List<string> files)
		{
			string dirPath = System.IO.Path.Combine(projectsFolder, projectName);

			// If we need to send a web request, do so.
			if (Application.streamingAssetsPath.Contains("://"))
			{
				LoadStatus dirFileLoadStatus = fileLoader.Load(dirPath);

				while (dirFileLoadStatus != LoadStatus.Loaded)
				{
					yield return null;
					dirFileLoadStatus = fileLoader.GetStatus(dirPath);
				}
				string dirIndexString = fileLoader.RetrieveResult(dirPath);

				files.AddRange(SplitIntoFileList(dirIndexString));
			}
			else
			{
				try
				{
					string fullPath = System.IO.Path.Combine(Application.streamingAssetsPath, dirPath);
					foreach (string file in System.IO.Directory.GetFiles(fullPath, "*.json"))
					{
						files.Add(System.IO.Path.GetFileName(file));
					}
				}
				catch (DirectoryNotFoundException e)
				{
					yield break;
				}
			}
		}

		IEnumerable<string> SplitIntoFileList(string dirIndexString)
		{
			// FIXME: TODO: implement
			throw new NotImplementedException();
		}

	}
}
