﻿using System;
using System.Collections.Generic;
using System.Linq;
using QuickGraph.Algorithms.ShortestPath;
using StudyCrafter.Assistant.Representation;

namespace StudyCrafter.Assistant.StudyCrafterScenario
{
    public static class InteractionMapExtensions
    {
        public static IEnumerable<InteractionNode> FindEventBasedInteractionNodes(this InteractionMap map)
        {
            HashSet<InteractionNode> result = new HashSet<InteractionNode>();
            foreach (InteractionNode node in map.Graph.Vertices)
            {
                if (node.NodeType == InteractionNode.Type.InteractionPoint)
                {
                    StringProperty scriptNodeType;
                    if (!node.TryGetProperty("script_type", out scriptNodeType))
                    {
                        continue;
                    }

                    if (scriptNodeType.Val == ScriptNodeTypes.Event)
                    {
                        result.Add(node);
                    }
                }
            }

            return result;
        }

        public static IEnumerable<InteractionNode> FindEndNodes(this InteractionMap map)
        {
            return map.Graph.Vertices.Where((node) => node.NodeType == InteractionNode.Type.End);
        }

        public static bool HasPathToAtLeastOne(this InteractionMap map, InteractionNode node, IEnumerable<InteractionNode> endNodes)
        {
            if (node.NodeType == InteractionNode.Type.End)
            {
                return true;
            }

            DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge> pathAlg = new DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge>(map.Graph, AllEdgesAvailable);
            pathAlg.Compute(node);

            bool foundEnd = false;
            foreach (InteractionNode endNode in endNodes)
            {
                double dist;
                if (pathAlg.TryGetDistance(endNode, out dist) && dist < double.MaxValue - 1)
                {
                    foundEnd = true;
                }
            }

            return foundEnd;
        }

        public static IEnumerable<InteractionNode> FindDeadEnds(this InteractionMap map)
        {
            HashSet<InteractionNode> result = new HashSet<InteractionNode>();
            foreach (InteractionNode node in map.Graph.Vertices)
            {
                if (node.NodeType == InteractionNode.Type.End)
                {
                    continue;
                }

                IEnumerable<InteractionEdge> edges;
                if (!map.Graph.TryGetOutEdges(node, out edges) ||
                   edges.Count() == 0)
                {
                    result.Add(node);
                }
            }

            return result;
        }

        public static InteractionNode GetStart(this InteractionMap graph)
        {
            InteractionNode startVertex;
            if (!graph.TryGetVertex(new InteractionNode("0", InteractionNode.Type.Start), out startVertex))
            {
                throw new NullReferenceException("No start vertex in interaction map.");
            }
            return startVertex;
        }

        public static InteractionNode GetFurthestNode(this InteractionMap map, DijkstraShortestPathAlgorithm<InteractionNode, InteractionEdge> shortestPathAlg)
        {
            InteractionNode result = null;
            shortestPathAlg.TryGetRootVertex(out result);
            double greatestDistance = 0;
            foreach (InteractionNode node in map.Graph.Vertices)
            {
                double distance;
                // We want the furthest node, but a reachable one. Distances at double.MaxValue are not reachable.
                if (shortestPathAlg.TryGetDistance(node, out distance) && !Double.IsPositiveInfinity(distance) && distance < double.MaxValue - 1)
                {
                    if (distance > greatestDistance)
                    {
                        greatestDistance = distance;
                        result = node;
                    }
                }
            }
            return result;
        }

        // Path distance calculation methods:

        public static double AllEdgesAvailable(InteractionEdge edge)
        {
            return 1;
        }

        public static double StopAtInteractionPoints(InteractionEdge edge)
        {
            if (edge.Target.NodeType == InteractionNode.Type.InteractionPoint)
            {
                return Double.PositiveInfinity;
            }
            else
            {
                return 1;
            }
        }
    }
}