﻿
using System;
using System.Collections.Generic;
using StudyCrafter.Assistant.StudyCrafterSaving;
using Newtonsoft.Json;
using StudyCrafter.Assistant.Loading;
using StudyCrafter.Assistant.Metrics;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterScenario;
using UnityEngine;
using StudyCrafter.Assistant.StudyCrafterLoading;

namespace StudyCrafter.Assistant.UI
{
	[RequireComponent(typeof(StreamingAssetsProjectLoader))]
	public class ProjectAnalysisComponent : MonoBehaviour
	{
		[SerializeField]
		private string projectName = null;

		[SerializeField]
		private string scriptNodeDefsPath = "ScriptNodeDefs";

		[SerializeField]
		private string dotExportFilepath = "";

		[SerializeField]
		private string metricsExportFilepath = "";

		public string RepresentationExportPath
		{
			get
			{
				return dotExportFilepath;
			}
			set
			{
				dotExportFilepath = value;
			}
		}

		public string MetricsExportPath
		{
			get
			{
				return metricsExportFilepath;
			}
			set
			{
				metricsExportFilepath = value;
			}
		}

		public string ProjectsDirectoryPath
		{
			get
			{
				return projectLoader.ProjectsFolder;
			}
			set
			{
				projectLoader.ProjectsFolder = value;
			}
		}

		private IProjectLoader projectLoader;

		private ScenarioLoader scenarioLoader = new ScenarioLoader();

		private IGraphStore graphStore = new GraphStore();

		private IGraphIDGenerator graphIDGenerator = new SequentialIDGenerator();

		private ScriptNodeDefCollection scriptNodeDefs = null;

		private Scenario scenario;

		private ProjectLoadStatus loadStatus;

		private MetricsCollator metricsCollator;

		private bool isIdle = true;

		public delegate void FinishedEvent();

		public event FinishedEvent FinishedEventHandler;

		public delegate void ErrorEvent(string error);

		public event ErrorEvent ErrorEventHandler;

		public ProjectAnalysisComponent()
		{
			InitMetrics();
		}

		private IDictionary<string, IScenarioMetric> InitMetrics()
		{
			return new Dictionary<string, IScenarioMetric>()
			{
				{ "connected_components", new NumConnectedComponents(graphStore) },
				{ "avg_out_degree", new AvgOutDegree(graphStore) },
				{ "total_script_nodes", new TotalScriptNodesMetric(graphStore) },
				{ "average_script_nodes", new AverageScriptNodesMetric(graphStore) },
				{ "total_layout_nodes", new TotalLayoutNodesMetric(graphStore) },
				{ "average_layout_nodes", new AverageLayoutNodesMetric(graphStore) },
				{ "num_scenes", new NumScenesMetric(graphStore) },
				{ "average_num_interaction_points", new AverageNumInteractionPoints(graphStore) },
				{ "average_num_choices_per_scene", new AverageNumChoicesPerScene(graphStore) },
				{ "num_choices_per_interaction_point", new ChoicesPerInteractionPoint(graphStore) },
				{ "feedback_per_interaction_unit", new FeedbackPerInteraction(graphStore) },
				{ "feedback_per_choice", new FeedbackPerChoice(graphStore) },
				{ "feedback_to_event_ratio", new FeedbackEventRatio(graphStore) },
				{ "text_chars_per_dialog_node", new TextPerDialogNode(graphStore) },
				{ "average_actions_taken", new AverageActionsTaken(graphStore) },
				{ "actions_taken_variance", new ActionsTakenVariance(graphStore) },
				{ "average_edges_traversed", new AverageEdgesTraversed(graphStore) },
				{ "edges_traversed_variance", new EdgesTraversedVariance(graphStore) },
				{ "intra_playthrough_diversity", new IntraPlaythroughDiversity(graphStore) },
				{ "global_diversity", new GlobalDiversity(graphStore) },
				{ "renewal_rate", new RenewalRate(graphStore) },
				{ "choice_range", new ChoiceRange(graphStore) },
				{ "choice_frequency", new ChoiceFrequency(graphStore) },
				{ "choice_variability", new ChoiceVariability(graphStore) }
			};
		}

		public void Awake()
		{
			projectLoader = GetComponent<StreamingAssetsProjectLoader>();
		}

		public void Start()
		{
			Debug.Log("Loading...");

			LoadScriptNodeDefs();
		}

		private void LoadScriptNodeDefs()
		{
			// FIXME: Should this be loaded from streaming assets instead?
			TextAsset scriptNodeDefsFile = Resources.Load(scriptNodeDefsPath) as TextAsset;

			scriptNodeDefs = JsonConvert.DeserializeObject<ScriptNodeDefCollection>(scriptNodeDefsFile.text);
		}

		public void Update()
		{
			if (null != loadStatus)
			{
				if (loadStatus.Status == LoadStatus.Loaded)
				{
					OnLoaded(projectName);
					loadStatus.UpdateStatus(LoadStatus.None);
				}
				else if (loadStatus.Status == LoadStatus.Error)
				{
					NotifyError("Failed to load project: " + projectName + ": " + loadStatus.ErrorText);
					isIdle = true;
				}
			}
		}

		private void OnLoaded(string project)
		{
			JsonScenarioStrings loadedProject = projectLoader.GetLoadedProject(project);
			StudyCrafterLoadHelpers loadHelpers = new StudyCrafterLoadHelpers(graphStore, graphIDGenerator, scriptNodeDefs);
			try
			{
				scenario = scenarioLoader.Load(loadedProject, loadHelpers);
			}
			catch (Exception e)
			{
				Debug.LogException(e);
			}

			Debug.Log("Loaded:" + project);

			try
			{
				ExportGraphs();
			}
			catch (Exception e)
			{
				isIdle = true;
				NotifyError("Could not write representation graphs: " + e.Message);
				return;
			}

			AnalyzeMetrics();
		}

		private void AnalyzeMetrics()
		{
			IDictionary<string, IScenarioMetric> metrics = InitMetrics();

			metricsCollator = new MetricsCollator();
			foreach (KeyValuePair<string, IScenarioMetric> metric in metrics)
			{
				metricsCollator.RegisterMetric(metric.Key, metric.Value);
			}
			metricsCollator.Finished += OnMetricsFinished;
			metricsCollator.Run(scenario);
		}

		private void ExportGraphs()
		{
			string filePrefix = scenario.ID.Trim().Replace(" ", "-");

			SceneGraph sceneGraph = scenario.GetSceneGraph(graphStore);
			foreach (Scene scene in sceneGraph.GetScenes())
			{
				InteractionMap interactionMap = scene.GetInteractionMap(graphStore);
				InteractionMapToDotExporter interactionExporter = new InteractionMapToDotExporter(dotExportFilepath, filePrefix, interactionMap);
				interactionExporter.Compute();

				ScriptGraph scriptGraph = scene.GetScriptGraph(graphStore);
				ScriptGraphToDotExporter scriptExporter = new ScriptGraphToDotExporter(dotExportFilepath, filePrefix, scriptGraph);
				scriptExporter.Compute();
			}
		}

		private void OnMetricsFinished()
		{
			try
			{
				MetricsCSVWriter writer = new MetricsCSVWriter();
				writer.Write(metricsExportFilepath, new Dictionary<string, MetricResultsCollection>() { { scenario.ID, metricsCollator.Results } });
			}
			catch (Exception e)
			{
				isIdle = true;
				NotifyError("Could not write results: " + e.Message);
				return;
			}

			isIdle = true;
			NotifyFinished();
		}

		private void NotifyFinished()
		{
			if (FinishedEventHandler != null)
			{
				FinishedEventHandler();
			}
		}

		private void NotifyError(string error)
		{
			if (ErrorEventHandler != null)
			{
				ErrorEventHandler(error);
			}
		}

		public void SetProject(string project)
		{
			projectName = project;
		}

		public void LoadAndAnalyze()
		{
			if (!isIdle)
			{
				Debug.LogError("Error: cannot analyze a new project - already in progress.");
				NotifyError("Analysis already in progress.");
				return;
			}

			isIdle = false;
			try
			{
				loadStatus = projectLoader.LoadProject(projectName);

				if (loadStatus.Status == LoadStatus.Error)
				{
					NotifyError("Failed to load project: " + projectName + ": " + loadStatus.ErrorText);
					isIdle = true;
				}
			}
			catch (Exception e)
			{
				NotifyError("Failed to load project: " + projectName + ": " + e.Message);
				Debug.LogException(e);
				isIdle = true;
			}
		}

		public bool IsReady()
		{
			return isIdle;
		}
	}
}
