﻿using System;
using System.Collections.Generic;
using StudyCrafter.Assistant.Metrics;
using StudyCrafter.Assistant.Representation;
using StudyCrafter.Assistant.StudyCrafterLoading;
using UnityEngine;

namespace StudyCrafter.Assistant.UI
{
	[RequireComponent(typeof(ProjectDirectoryLoader))]
	public class ScenarioDirectoryAnalyzerComponent : MonoBehaviour
	{
		[SerializeField]
		private string metricsOutputPath;

		private ProjectDirectoryLoader loader;
		private MultiScenarioMetricsAnalyzer analyzer;

		private IDictionary<string, IScenarioMetric> metrics;

		private bool isIdle = true;

		public string MetricsExportPath
		{
			get
			{
				return metricsOutputPath;
			}
			set
			{
				metricsOutputPath = value;
			}
		}

		public string ProjectsDirectoryPath
		{
			get
			{
				return loader.ProjectsFolder;
			}
			set
			{
				loader.ProjectsFolder = value;
			}
		}

		private IGraphStore GraphStore
		{
			get
			{
				return loader.GraphStore;
			}
		}

		public MetricsFinishedEvent AnalysisFinishedHandler;

		public delegate void ErrorEvent(string error);

		public event ErrorEvent ErrorEventHandler;

		public ScenarioDirectoryAnalyzerComponent()
		{
		}

		private IDictionary<string, IScenarioMetric> InitMetrics()
		{
			return new Dictionary<string, IScenarioMetric>()
			{
				{ "connected_components", new NumConnectedComponents(GraphStore) },
				{ "avg_out_degree", new AvgOutDegree(GraphStore) },
				{ "total_script_nodes", new TotalScriptNodesMetric(GraphStore) },
				{ "average_script_nodes", new AverageScriptNodesMetric(GraphStore) },
				{ "total_layout_nodes", new TotalLayoutNodesMetric(GraphStore) },
				{ "average_layout_nodes", new AverageLayoutNodesMetric(GraphStore) },
				{ "num_scenes", new NumScenesMetric(GraphStore) },
				{ "average_num_interaction_points", new AverageNumInteractionPoints(GraphStore) },
				{ "average_num_choices_per_scene", new AverageNumChoicesPerScene(GraphStore) },
				{ "num_choices_per_interaction_point", new ChoicesPerInteractionPoint(GraphStore) },
				{ "feedback_per_interaction_unit", new FeedbackPerInteraction(GraphStore) },
				{ "feedback_per_choice", new FeedbackPerChoice(GraphStore) },
				{ "feedback_to_event_ratio", new FeedbackEventRatio(GraphStore) },
				{ "text_chars_per_dialog_node", new TextPerDialogNode(GraphStore) },
				{ "average_actions_taken", new AverageActionsTaken(GraphStore) },
				{ "actions_taken_variance", new ActionsTakenVariance(GraphStore) },
				{ "average_edges_traversed", new AverageEdgesTraversed(GraphStore) },
				{ "edges_traversed_variance", new EdgesTraversedVariance(GraphStore) },
				{ "intra_playthrough_diversity", new IntraPlaythroughDiversity(GraphStore) },
				{ "global_diversity", new GlobalDiversity(GraphStore) },
				{ "renewal_rate", new RenewalRate(GraphStore) },
				{ "choice_range", new ChoiceRange(GraphStore) },
				{ "choice_frequency", new ChoiceFrequency(GraphStore) },
				{ "choice_variability", new ChoiceVariability(GraphStore) }
			};
		}

		public void Awake()
		{
			loader = GetComponent<ProjectDirectoryLoader>();
			loader.Complete += OnLoaded;
			loader.ErrorHandler += OnError;
		}

		public void Start()
		{
		}

		private void OnLoaded()
		{
			analyzer = new MultiScenarioMetricsAnalyzer(() => InitMetrics());

			analyzer.Finished += OnAnalysisFinished;

			Debug.Log("Analyzing metrics!");

			analyzer.AnalyzeMetrics(loader.Scenarios.Values);
		}

		private void OnAnalysisFinished()
		{
			Debug.Log("Done analyzing - saving metrics.");

			try
			{
				MetricsCSVWriter writer = new MetricsCSVWriter();
				writer.Write(metricsOutputPath, analyzer.Results);
			}
			catch (Exception e)
			{
				OnError("Could not write results: " + e.Message);
				return;
			}

			analyzer.Finished -= OnAnalysisFinished;

			Debug.Log("Metrics saving finished!");

			isIdle = true;

			NotifyFinished();
		}

		private void OnError(string message)
		{
			isIdle = true;
			NotifyError(message);
		}

		private void NotifyFinished()
		{
			if (null != AnalysisFinishedHandler)
			{
				AnalysisFinishedHandler();
			}
		}

		private void NotifyError(string message)
		{
			if (null != ErrorEventHandler)
			{
				ErrorEventHandler(message);
			}
		}

		public void Analyze()
		{
			if (isIdle == false)
			{
				Debug.LogError("Analysis already in progress. Can't start again.");
				return;
			}
			isIdle = false;
			loader.LoadProjects();
		}
	}
}
